#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#include <assert.h>
#include <math.h>
#include <segmentation.h>

#define MIN3(a, b, c) ( (a) < (b) ? ( (a) < (c) ? (a) : ( (b) < (c) ? (b) : (c) ) ) : ( (b) < (c) ? (b) : (c) ) )
#define MAX3(a, b, c) ( (a) > (b) ? ( (a) > (c) ? (a) : ( (b) > (c) ? (b) : (c) ) ) : ( (b) > (c) ? (b) : (c) ) )

static void  _makeComponentConnections(int * connections, int * clusters, int nbColors);

void byComponentHistogram(uint8_t * data, int w, int h, int ws, int nbc, int * histogram, int shift) {
  int i, j, ii;
  for(i = 0; i < h; i++) {
    ii = i * ws;
    for(j = 0; j < w; j++)
      histogram[data[ii + j * nbc + shift]]++;
  }
}

void segmentUsingKmean(uint8_t * data, int w, int h, int ws, int nbc, int nbColors, int shift) {
  uint64_t * m = NULL, * nm = NULL;
  int * clusters = NULL, * histogram = NULL, i, ii, j, k, maxv, next, end = 0, closest, connections[256];
  histogram = calloc(256, sizeof *histogram); assert(histogram);
byComponentHistogram(data, w, h, ws,  nbc, histogram, shift);
  clusters = malloc(nbColors * sizeof *clusters); assert(clusters);
  m = malloc(nbColors * sizeof *m); assert(m);
  nm = malloc(nbColors * sizeof *nm); assert(nm);
  /* sélection des nbColors meilleures valeurs dans l'histo et
   * transfert vers le tableau clusters */
  for(j = 0, maxv = -1; j < 256; j++) {
    if(maxv < histogram[j]) {
      maxv = histogram[j];
      clusters[0] = j;
    }
  }
  for(i = 1; i < nbColors; i++) {
    for(j = 0, maxv = -1; j < 256; j++) {
      for(k = i - 1, next = 0; k >= 0; k--) {
	if(j == clusters[k]) {
	  next = 1;
	  break;
	}
      }
      if(next == 0 && maxv < histogram[j]) {
	maxv = histogram[j];
	clusters[i] = j;
      }
    }
  }
  /* convergeance */
  while(!end) {
    /* attribution des luminances aux clusters */
    _makeComponentConnections(connections, clusters, nbColors);
    for(i = 0; i < nbColors; i++)
      m[i] = (nm[i] = 0);
    for(i = 0; i < 256; i++) {
      m[closest = connections[i]] += i * histogram[i];
      nm[closest] += histogram[i];
    }
    end = 1;
    for(i = 0; i < nbColors; i++) {
      if(nm[i] > 0) m[i] /= nm[i];
      if(clusters[i] != m[i]) {
	end = 0;
	clusters[i] = (int)m[i];
      }
    }
  }
  for(i = 0; i < h; i++) {
    ii = i * ws;
    for(j = 0; j < w; j++) {
      k = ii + j * nbc + shift;
      data[k] = clusters[connections[data[k]]];
    }
  }
  free(histogram);
  free(clusters);
  free(m);
  free(nm);
}

#define UNDEFINED 1.0

void rgb2hsv(double r, double g, double b, double * h, double * s, double * v) {
  double min = MIN3(r, g, b), max = MAX3(r, g, b);
  *v = max;
  *s = (max != 0.0) ? ((max - min) / max) : 0.0;
  if(*s == 0.0)
    *h = UNDEFINED;
  else {
    double delta = 1.0 / (max - min);
    if(r == max)
      *h = delta * (g - b) ;
    else if(g == max)
      *h = 2.0 + delta * (b - r);
    else if(b == max)
      *h = 4.0 + delta * (r - g);
    *h = (*h < 0.0) ? (*h / 6.0 + 1.0) : (*h / 6.0);
  }
}

void hsv2rgb(double h, double s, double v, double * r, double * g, double * b) {
  if(s == 0.0) {
    if(h == UNDEFINED)
      *r = *g = *b = v;
    else
      return;
  } else {
    int i;
    double f, p, q, t;
    i = h = (h == 1.0) ? 0.0 : 6.0 * h;
    f = h - i;
    p = v * (1.0 - s);
    q = v * (1.0 - s * f);
    t = v * (1.0 - s * (1.0 - f));
    switch(i) {
    case 0:  *r = v; *g = t; *b = p; break;
    case 1:  *r = q; *g = v; *b = p; break;
    case 2:  *r = p; *g = v; *b = t; break;
    case 3:  *r = p; *g = q; *b = v; break;
    case 4:  *r = t; *g = p; *b = v; break;
    default: *r = v; *g = p; *b = q; break;
    }
  }
}

#undef UNDEFINED

static void _makeComponentConnections(int * connections, int * clusters, int nbColors) {
  int i, j, diff, closest;
  for(i = 0; i < 256; i++) {
    diff = abs(i - clusters[closest = 0]);
    for(j = 1; j < nbColors; j++)
      if(diff > abs(i - clusters[j]))
	diff = abs(i - clusters[closest = j]);
    connections[i] = closest;
  }
}

